/*
    $Id: CButlerPhone.cc 78 2009-05-29 15:41:03Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include <pthread.h>
#include <time.h>
#include <errno.h>

#include "CLog.h"
#include "CButlerPhone.h"
#include "CButlerPhoneClient.h"

CButlerPhone::CButlerPhone(CButlerPhoneClient *owner) {
    this->owner = owner;

    lastCommand = ' ';

    isConnected = false;
    statusButler = BSClearDisplay;
    statusCalls = CSNoCalls;

    threadConnect = NULL;
    threadSend = NULL;
    threadStatus = NULL;
    threadRinger = NULL;

    pthread_mutex_init(&synchro_mutex, NULL);
    
    fifo = new CSendFifo();
}

CButlerPhone::~CButlerPhone(void) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone: Destroying CButlerPhone");
    if (isConnected) {
        if (threadStatus) {
            CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone: Connected: terminating status thread");
            threadStatus->terminate();
            delete threadStatus;
        }

        CLog::Log(LOGLEVEL_INFO,"CButlerPhone: Stop ringing");
        stopRing();

        CLog::Log(LOGLEVEL_INFO,"CButlerPhone: Hangup");
        hangup();

        CLog::Log(LOGLEVEL_INFO,"CButlerPhone: Clear display");
        setButlerStatus(BSClearDisplay);
        setCallsStatus(CSNoCalls);

        CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone: Waiting for send fifo empty");
        while (!fifo->isEmpty()) {
            sleep(1);
        }

        if (threadSend) {
            CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone: Terminating send thread");       
            threadSend->terminate();
            delete threadSend;
       }

        CLog::Log(LOGLEVEL_INFO,"CButlerPhone: Reset and close device");
        usb_reset(h_USBDevice);
        usb_release_interface(h_USBDevice,0);
        usb_close(h_USBDevice);
    }
    else {
        CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone: Not connected: terminating connect thread");
        threadConnect->terminate();
        delete threadConnect;
    }

    CLog::Log(LOGLEVEL_INFO,"CButlerPhone: Destroying send fifo");
    delete fifo;
    fifo = NULL;
}

void CButlerPhone::start(void) {
    if (!isConnected && (threadConnect==NULL))
        threadConnect = new CConnectThread(this);
}

void CButlerPhone::set_Connected(bool newStatus, usb_dev_handle *h_USBDevice) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::set_Connected");
    if (newStatus == isConnected)
        return;

    isConnected = newStatus;

    if (isConnected) {
        CLog::Log(LOGLEVEL_INFO,"CButlerPhone::set_Connected: connected");
        //threadConnect->terminate();
        //delete threadConnect;
        // TODO: wann kann threadConnect geloescht werden?
        threadConnect = NULL;

        this->h_USBDevice = h_USBDevice;

        if (threadStatus) {
            threadStatus->terminate();
            delete threadStatus;
            threadStatus = NULL;
        }

        if (threadSend) {
            threadSend->terminate();
            delete threadSend;
            threadSend = NULL;
        }

        threadSend = new CSendThread(h_USBDevice, fifo);
        threadStatus = new CStatusThread(this, h_USBDevice);

        sendStatus();
        owner->onButlerConnected();
    }
    else {
        CLog::Log(LOGLEVEL_INFO,"CButlerPhone::set_Connected: disconnected");

        owner->onButlerDisconnected();

        if (threadRinger != NULL) {
            threadRinger->terminate();
            delete threadRinger;
            threadRinger = NULL;
        }

        fifo->clear();

        usb_reset(this->h_USBDevice);
        usb_release_interface(this->h_USBDevice, 0);
        usb_close(this->h_USBDevice);
        this->h_USBDevice = NULL;

        if (threadConnect) {
            threadConnect->terminate();
            delete threadConnect;
            threadConnect = NULL;
        }

        threadConnect = new CConnectThread(this);
    }
}

void CButlerPhone::receive(const uchar *s, int size) {
    CLog::LogHex(LOGLEVEL_DEBUG,"CButlerPhone::receive",s, size);
    if ((s == NULL) || (size < 1))
        return;

    switch (s[0]) {
        //------- 7E : are you there?
        case 0x7e:  send(&s[0],1); //yes I am
            break;

        //------- 23 : parameters of the phone (reply to 85 01)
        //             Ignored since I don't know what they mean
        case 0x23:  break;

        //------- 24 : you didn't reply, I'm blanking
        case 0x24:  sendStatus();
            break;

        //------- 50 : contacts keys
        case 0x50:  owner->onButlerKeyContacts(getKey(s,size));
        break;

        //------- 53 : calls keys
        case 0x53:  owner->onButlerKeyCalls(getKey(s,size));
        break;

        //------- 55 : calls exit
        case 0x55: 
            owner->onButlerKeyCalls(CKExit);
            sendStatus();
        break;

        //------- 54 : redial keys
        case 0x54: owner->onButlerKeyRedial(getKey(s,size));
        break;

        //------- 30 01 x DD DD ... dialled digits
        //------- 30 00 x CC CC ... dialled ucharacters (data on display) ??
        //------- 30 00 x 01 CC CC ... dialled ucharacters (data on display)
        case 0x30:
            if (size < 4)
                return;
            if ((s[1] != 0x01) && (s[1] != 0x00))
                return;
            send(&s[0], 1);
            command((uchar[]){0x91, 0x01}, 2);

            char dialstr[s[2]+1]; // TODO check length
            if (s[1] == 0x01) {                
                decodeNumber(s[2], dialstr, &s[3], size-3);
                dialstr[s[2]] = '\0';
                owner->onButlerDialNumber(dialstr);
            }
            else {
                memcpy(dialstr, &s[4], size-4);
                dialstr[size-4] = '\0';
                owner->onButlerDialName(dialstr);
            }

        break;

        //------- 35 : silent ringing
        case 0x35: send(&s[0],1); //confirm and ignore it
        break;

        //------- 36 : accept call (after silent ringing)
        //------- 31 : accept call (after ringing)
        case 0x36:
        case 0x31:
            stopRing();
            send(&s[0], 1);
            command((uchar[]){0x91, 0x01}, 2);
            owner->onButlerAcceptCall();
        break;

        //------- 32 : hang up
        case 0x32:
            send(&s[0], 1);
            sendStatus(); //to avoid auto blinking of the calls icon
            owner->onButlerHangup();
        break;

        //------- 34 01 DD : dialled digit (during a call)
        case 0x34:
            if (size < 3)
                break;

            char decode;
            if (decodeNumber(1,&decode,&s[2],1))
                owner->onButlerDigit(&decode);
        break;

        default:
            if (s[0] == lastCommand) {
                CLog::Log(LOGLEVEL_DEBUG, "CButlerPhone::receive: AKN command %0x", s[0]);
            }
            else {
                CLog::Log(LOGLEVEL_INFO, "CButlerPhone::receive: Unhandled command %0x", s[0]);
            }

        break;
    }
}

void CButlerPhone::command(const uchar *s, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::command");
    if ((s == NULL) || (size < 1))
        return;

    lastCommand = s[0];
    send(s,size);
}

pthread_mutex_t *CButlerPhone::getMutex(void) {
    return &this->synchro_mutex;
}

void CButlerPhone::send(const uchar *s, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::send");
    if (!isConnected)
        return;
    CLog::LogHex(LOGLEVEL_DEBUG,"CButlerPhone::send: Message",s,size);
    fifo->push(s,size);
}

void CButlerPhone::sendStatus(void) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::sendStatus");
    uchar cmd1[] = {0x84, (uchar)statusButler};
    command(cmd1, 2);
    uchar cmd2[] = {
        0x8c,
        (uchar)statusButler,
        (uchar)statusCalls,
        0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
    };
    command(cmd2, 10);  
}

int CButlerPhone::decodeNumber(int digits, char *decode, const uchar *code, int sizeCode) {
    const char *codes = "0123456789*# P  ";
    int i,n;

    if (sizeCode < div(digits+1,2).quot) {
        return 0;
    }
    
    for (i=0; i < digits; i++) {
        n = div(i,2).quot;
        if (n*2 == i) 
            decode[i] = codes[code[n] & 0x0f];
        else
            decode[i] = codes[(code[n] & 0xf0) >> 4];
    }

    return i;
}

uchar CButlerPhone::checksum(uchar *string, int size) {
    int i;
    uchar c=0;
 
    for (i = 0; i < size; i++) c = c ^ string[i];
  
    return c;
}

CButlerPhone::contactKey_t CButlerPhone::getKey(const uchar *s, int size) {
    if (size < 4)
        return CKUndef;

    // Unified key processing:
    //   XX FF FF 00 : XX open
    //   XX 00 00 00 : XX down
    //   XX 00 00 01 : XX up
    if (s[1] == 0xff)
        return CKFirst;
    else if (s[3] == 0x00)
        return CKDown;
    else
        return CKUp;
}

void CButlerPhone::setButlerStatus(CButlerPhone::TButlerStatus newStatus) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::setButlerStatus");
    if (newStatus == statusButler)
        return;

    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::setButlerStatus old=%u new=%u", statusButler, newStatus);
    statusButler = newStatus;
    sendStatus();
}

CButlerPhone::TButlerStatus CButlerPhone::getButlerStatus() {
    return statusButler;
}

void CButlerPhone::setCallsStatus(CButlerPhone::TCallsStatus newStatus) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::setCallsStatus");
    if (newStatus == statusCalls)
        return;

    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::setCallsStatus old=%u new=%u", statusCalls, newStatus);
    statusCalls = newStatus;
    sendStatus();
}

CButlerPhone::TCallsStatus CButlerPhone::getCallsStatus() {
    return statusCalls;
}

void CButlerPhone::displayContact(const char *s, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::displayContact");
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+5];
    uchar cmd1[] = {0x86, 0x01, 0x00, size, 0x01};
    memcpy(buf,cmd1,5*sizeof(uchar));
    memcpy(&buf[5],(uchar*)s,size*sizeof(uchar));
    command(buf,size+5);
    delete [] buf;
}

void CButlerPhone::displayId(const char *id, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::displayId");
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+2];
    uchar cmd1[] = {0x87, size};
    memcpy(buf,cmd1,2*sizeof(uchar));
    memcpy(&buf[2],(uchar*)id,size*sizeof(uchar));
    command(buf,size+2);
    delete [] buf;
}

void CButlerPhone::displayCall(const char *s, int size,  bool isOnline, time_t dt) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::displayCall");
    tm *datetime = localtime(&dt);
    uchar chonline = (isOnline) ? 0x10 : 0x01;    
    if (datetime == NULL) {
        CLog::Log(LOGLEVEL_INFO, "displayCall cannot decode time");
        return;
    }
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+12];
    uchar cmd1[] = {0x89, 0x00, 0x00, 
        chonline, 
        datetime->tm_year, datetime->tm_mon + 1, datetime->tm_mday, 
        datetime->tm_hour, datetime->tm_min, datetime->tm_sec,
        size, 0x01
    };
    memcpy(buf,cmd1,12*sizeof(uchar));
    memcpy(&buf[12],(uchar*)s,size*sizeof(uchar));
    command(buf,size+12);
    delete [] buf;
}

void CButlerPhone::displayRedial(const char *s, int size,  bool isOnline, time_t dt) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::displayRedial");
    tm *datetime = localtime(&dt);
    uchar chonline = (isOnline) ? 0x10 : 0x00;    
    if (datetime == NULL) {
        CLog::Log(LOGLEVEL_INFO, "displayCall cannot decode time");
        return;
    }
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+12];
    uchar cmd1[] = {0x8a, 0x00, 0x00, 
        chonline, 
        datetime->tm_year, datetime->tm_mon, datetime->tm_mday, 
        datetime->tm_hour, datetime->tm_min, datetime->tm_sec,
        size, 0x01
    };
    memcpy(buf,cmd1,12*sizeof(uchar));
    memcpy(&buf[12], (uchar*)s, size*sizeof(uchar));
    command(buf, size+12);
    delete [] buf;
}

void CButlerPhone::ring(const char *callerid, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::ring");
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+4];
    uchar cmd1[] = {0x81, size, 0x01};
    memcpy(buf,cmd1, 3*sizeof(uchar));
    memcpy(&buf[3], (uchar*)callerid, size*sizeof(uchar));
    buf[size+3] = 0x01;
    command((uchar[]){0x80, 0x01}, 2);
    command(buf, size+4);
    delete [] buf;

    if (threadRinger == NULL)
        threadRinger = new CRingerThread(this);
}

void CButlerPhone::silentRing(const char *callerid, int size) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::silentRing");
    if (size>20)
        size = 20;
    uchar *buf = new uchar[size+4];
    uchar cmd1[] = {0x90,size,0x00};
    memcpy(buf, cmd1, 3*sizeof(uchar));
    memcpy(&buf[3], (uchar*)callerid, size*sizeof(uchar));
    buf[size+3] = 0x01;
    command(buf, size+4);
    delete [] buf;
}

void CButlerPhone::stopRing(void) {
    CLog::Log(LOGLEVEL_DEBUG,"CButlerPhone::stopRing");
    if (threadRinger != NULL) {
        threadRinger->terminate();
        delete threadRinger;
        threadRinger = NULL;
    }
    command((uchar[]){0x80,0x00}, 2);
}

void CButlerPhone::hangup(void) {
    command((uchar[]){0x82,0x82}, 2);
    sendStatus();   
}

