/*
    $Id: CButlerPhone.h 54 2009-03-30 23:44:18Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __CBUTLERPHONE_H__
#define __CBUTLERPHONE_H__

#include <usb.h>
#include <time.h>
#include <deque>

//#include "CButlerPhoneClient.h"

#define uchar unsigned char

class CButlerPhoneClient;

class CButlerPhone {

public:
    // The status is displayed when the phone is idle in the selected
    // language
    typedef enum {
        BSOffline, //keypresses aren't reported to the pc in this state
        BSOnline,
        BSSkypeMe,
        BSAway,
        BSNotAvailable,
        BSDoNotDisturb,
        BSInvisible,
        BSLogOff,
        //most features of the phone don't work in the following
        //states (e.g. caller id display, calls status icon)
        BSClearDisplay=16,
        BSMSN=17,
        BSSJPhone=33,
        BSNet2Phone=49
    } TButlerStatus;

    // Calls icon (only works if in state BSOffline..BSLogOff)
    typedef enum {
        CSNoCalls,   // icon off
        CSCalls,     // icon on
        CSNewCalls // icon blinking
    } TCallsStatus;

    //Key for the "calls", "redial" and "contacts" events
    //CKExit is only fired by the "calls" key
    typedef enum {
        CKUndef=-1,
        CKFirst,
        CKUp,
        CKDown,
        CKExit
    } contactKey_t;

private:
    typedef void(*t_NotifyEvent)(void *sender);
    typedef void(*t_NotifyKey)(void *sender, contactKey_t key);

    class CRingerThread;
    class CStatusThread;
    class CSendThread;
    class CSendFifo;
    class CConnectThread;

    CButlerPhoneClient *owner;

    usb_dev_handle *h_USBDevice;

    TButlerStatus statusButler;
    TCallsStatus statusCalls;
    pthread_mutex_t synchro_mutex;

    CSendFifo *fifo;

    uchar lastCommand;
    bool isConnected;
    bool isIgnoreChecksum;

    CConnectThread *threadConnect;
    CSendThread *threadSend;
    CStatusThread *threadStatus;
    CRingerThread *threadRinger;

    void send(const uchar *s, int size);
    void receive(const uchar *s, int size);

    void sendStatus(void);

    void set_Connected(bool newStatus, usb_dev_handle *h_USBDevice);    
    static int decodeNumber(int digits, char *decode, const uchar *code, int sizeCode);
    static uchar checksum(uchar *string, int size);
    contactKey_t getKey(const uchar *s, int size);

    class CRingerThread {
        CButlerPhone *owner;
        void syncProc(void);
        void *execute(void);
        pthread_t thread;
        static void *threadfunc(void *ptr);

    public:
        CRingerThread(CButlerPhone *owner);
        void terminate(void);
    };

    class CStatusThread {
        CButlerPhone *owner;
        usb_dev_handle *h_USBDevice;
        uchar *data;
        int size;
        void receive(void);
        void error(void);
        void *execute(void);
        pthread_t thread;
        static void *threadfunc(void *ptr);

    public:
        CStatusThread(CButlerPhone *owner, usb_dev_handle *h_USBDevice);
        void terminate(void);
    };

    class CSendThread {
        usb_dev_handle *h_USBDevice;
        CSendFifo *sendFifo;
        pthread_mutex_t condition_mutex;
        pthread_cond_t  condition_cond;

        void *execute(void);
        pthread_t thread;
        static void *threadfunc(void *ptr);
        void signalWork(void);

    public:
        static void onWork(void *sender);
        CSendThread(usb_dev_handle *h_USBDevice, CSendFifo *sendFifo);
        void terminate(void);
    };

    class CSendFifo {
    
    public:
        typedef struct {
            int size;
            uchar *data;
        } fifoData_t;

    private:
        std::deque<fifoData_t> pipe;
        t_NotifyEvent onPush;
        void *onPush_arg;
        pthread_mutex_t critical_mutex;

    public:
        CSendFifo(void);
        void setHandlerPush(t_NotifyEvent handler, void *arg);
        void push(const uchar *string, int size);
        fifoData_t *pop(void);
        bool isEmpty(void);
        void clear(void);
        ~CSendFifo(void);
    };

    class CConnectThread {
        CButlerPhone *owner;
        usb_dev_handle *h_USBDevice;
        pthread_t thread;
        void syncProc(void);
        void *execute(void);
        static void *threadfunc(void *ptr);
        struct usb_device *locateButler(void);
        bool initButler(void);

    public:
        CConnectThread(CButlerPhone *owner);
        ~CConnectThread(void);
        void terminate(void); 
    };

protected:
    void command(const uchar *s, int size);
    pthread_mutex_t *getMutex(void);

public:
    CButlerPhone(CButlerPhoneClient *owner);
    ~CButlerPhone(void);

    //after you have assigned the various event handlers you shoul call this
    //method to start the communication with the phone
    void start(void);

    //this one only displays a name, it's used for the "contacts" events,
    //since in that mode it's not possible to see the details of a call
    void displayContact(const char *s, int size);

    //this one displays the name but can also display the date/time of
    //the call and the current state of the contact, so it's used for "calls"
    //events.
    void displayCall(const char *s, int size,  bool isOnline, time_t dt);

    //this one does exactly the same as the previous one but it's used for
    //the "redial" events, though they both work in each other context
    void displayRedial(const char *s, int size, bool isOnline, time_t dt);

    //this one displays the id of the contact after pressing the "Ok" key.
    //It doesn't overwrite the name buffer, so a subsequent "Ok" won't send
    //a message to the pc, and a "Dial" will use the name, not the id
    void displayId(const char *id, int size);

    //Rings the phone and displays the callerid (limited to 20 ucharacters)
    void ring(const char *callerid, int size);

    //Stops ringing
    void stopRing(void);

    //This method will only display the callerid without ringing the phone
    void silentRing(const char *callerid, int size);

    //This method will hangup the phone during a call, to be used when the
    //call is terminated by the other party or if you want to terminate
    //a call request coming from the phone (i.e. after an OnDialNumber or
    //OnDialName event)
    void hangup(void);

    //Changes the status displayed on the phone
    void setButlerStatus(TButlerStatus newStatus);
    TButlerStatus getButlerStatus(void);

    //Activates the calls icon
    void setCallsStatus(TCallsStatus newStatus);
    TCallsStatus getCallsStatus(void);

    //returns true if the phone is connected, false otherwise
    bool getConnected(void);
 
    //if a message received from the phone has a wrong checksum
    //ignore the checksum instead of discarding the message 
    void setIngoreChecksum(bool on);
    bool getIgnoreChecksum(void);       
};

#endif
