/*
    $Id: CConnectThread.cc 65 2009-05-03 19:07:03Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <usb.h>
#include <pthread.h>
#include <stdio.h>
#include <errno.h>

#include "CLog.h"
#include "CButlerPhone.h"

#define  DEV_VID 0xE6A
#define  DEV_PID 0x6002

CButlerPhone::CConnectThread::CConnectThread(CButlerPhone *owner) {
    this->owner = owner;
    h_USBDevice = NULL;

    if (0 != pthread_create( &thread, NULL, threadfunc, this))
        CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread: Create ConnectThread failed");
}

CButlerPhone::CConnectThread::~CConnectThread() {

    if (h_USBDevice)
        usb_close(h_USBDevice);
}

void *CButlerPhone::CConnectThread::threadfunc(void *ptr) {
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::threadfunc");
    return ((CConnectThread*)ptr)->execute();
}

void CButlerPhone::CConnectThread::terminate(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::terminate");

    pthread_cancel(thread);
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread: canceled");
    pthread_join(thread, NULL);
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread: terminated");

}

void CButlerPhone::CConnectThread::syncProc(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::syncProc");
    owner->set_Connected(true, h_USBDevice);
}

struct usb_device *CButlerPhone::CConnectThread::locateButler(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::locateButler");
    struct usb_bus *bus;
    struct usb_device *dev;
 		
    usb_find_busses();
    usb_find_devices();
 
    for (bus = usb_busses; bus; bus = bus->next) {
        for (dev = bus->devices; dev; dev = dev->next) {
            if ((dev->descriptor.idVendor == DEV_VID) && (dev->descriptor.idProduct == DEV_PID)) {	
                CLog::Log(LOGLEVEL_INFO, "Butler VoIP Device Found @ Address %s/%s", bus->dirname, dev->filename);
                return (dev);
            }
        }	
    }

    return NULL;  	
}

bool CButlerPhone::CConnectThread::initButler(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::initButler");
    uchar data[2];
    int send_status;

    if (h_USBDevice == NULL)
        return false;

    send_status = usb_control_msg(h_USBDevice, 0xc0, 0x04, 0x5004, 0x0f09, (char*)data, 2, 500); 
	CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::initButler: RX stat=%d data=0x%x,0x%x",send_status,data[0],data[1]);
    if (send_status != 2) {
        CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread::initButler: Init device failed initial read");
        return false;
    }

    data[0] = 0x90;
    send_status = usb_control_msg(h_USBDevice, 0x40, 0x0c, 0x5003, 0x00f0, (char*)data, 1 ,500);
    if (send_status != 1) {
        CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread::initButler: Init device failed send 0x90");
        return false;
    }

    data[0] = 0x08;
    send_status = usb_control_msg(h_USBDevice, 0x40, 0x0c, 0x5003, 0xffff, (char*)data, 1 ,500);
    if (send_status != 1) {
        CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread::initButler: Init device failed send 0x08");
        return false;
    }

    return true;
}
 
void *CButlerPhone::CConnectThread::execute(void) {
    struct usb_device *dev;
    CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::execute");
    pthread_mutex_t *synchro_mutex = owner->getMutex();

	usb_init();
	//usb_set_debug(2);

    while(1) {
        pthread_testcancel();
        if (!(dev = locateButler()))
            goto next;

        if (!(h_USBDevice = usb_open(dev))) {
            CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread::execute: Open device failed");
            goto next;
        }

        if (usb_claim_interface(h_USBDevice,0) < 0) {
            CLog::Log(LOGLEVEL_ERROR, errno, "CConnectThread::execute: Claim device failed");
            usb_close(h_USBDevice);
            h_USBDevice = NULL;
            goto next;
        }

        if (initButler()) {
            pthread_mutex_lock(synchro_mutex);
            pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, synchro_mutex);
            syncProc();
            //pthread_mutex_unlock(synchro_mutex);
            pthread_cleanup_pop(1);
            CLog::Log(LOGLEVEL_DEBUG, "CConnectThread::execute pthread_exit()");
            pthread_exit(NULL);
        }

next:
        sleep(1);
    }
}

