/*
    $Id: CSendThread.cc 67 2009-05-03 19:49:06Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <pthread.h>
#include <string.h>
#include <errno.h>

#include "CLog.h"
#include "CButlerPhone.h"

CButlerPhone::CSendThread::CSendThread(usb_dev_handle *h_USBDevice, CSendFifo *sendFifo) {
    this->h_USBDevice = h_USBDevice;
    this->sendFifo = sendFifo;

    pthread_mutex_init(&condition_mutex, NULL);
    pthread_cond_init(&condition_cond, NULL);

    sendFifo->setHandlerPush((t_NotifyEvent)&onWork, this);
    
    if (0 != pthread_create( &thread, NULL, threadfunc, this))
        CLog::Log(LOGLEVEL_ERROR, errno, "CSendThread: Create SendThread failed");
}

void *CButlerPhone::CSendThread::threadfunc(void *ptr) {
    return ((CSendThread*)ptr)->execute();
}

void *CButlerPhone::CSendThread::execute(void) {
   CLog::Log(LOGLEVEL_DEBUG,"CSendThread::execute");
    int size;
    CSendFifo::fifoData_t *fifodata;
    uchar *buf, cs, sendmsg[8];
    int i, i2, len;

    while (1) {
        while (1) {
            pthread_testcancel();
            if (sendFifo->isEmpty()) {
                CLog::Log(LOGLEVEL_DEBUG,"CSendThread::execute sendfifo empty cond_wait...");
                pthread_mutex_lock( &condition_mutex );
                pthread_cond_wait( &condition_cond, &condition_mutex );
                pthread_mutex_unlock( &condition_mutex );
                CLog::Log(LOGLEVEL_DEBUG,"CSendThread::execute end cond_wait");
            }

            if ((fifodata = sendFifo->pop()))
                break;
        }
        pthread_testcancel();

        size = fifodata->size;
        cs = checksum(fifodata->data, size);
        buf = new uchar[size+3];
        uchar cmd1[] = {0xf0,size+1};
        memcpy(buf, cmd1, 2*sizeof(uchar));
        memcpy(&buf[2], fifodata->data, size*sizeof(uchar));
        memcpy(&buf[2+size], &cs, sizeof(uchar));

        delete [] fifodata->data;
        delete fifodata;

        size += 3;
        // Now that we have a standard message (F0 nn dd dd ... cs) we have to send it.
        // Messages are sent with an interrupt write, but in each call only and always
        // 8 bytes are sent, the first one is '8x' (where x=1..7 is the number of
        // data bytes), the rest are the data bytes (padded with 0s if there are less
        // than 7)
        i=0; i2=0; len=0;

        while (size) {
            len = size;
            if (len > 7) len = 7;
            size -= len;

            sendmsg[0] = 0x80 + len;

            for (i=1; i <= len; i++) {
                sendmsg[i] = buf[i2];
                i2++;
            }
        
            if ( len < 7) {
                for (i=len+1; i < 8; i++) sendmsg[i] = 0x0;
            }
            CLog::LogHex(LOGLEVEL_DEBUG,"CSendThread::execute: Sending",sendmsg, 8);

            int sent = usb_interrupt_write(h_USBDevice, 3, (char *)sendmsg, 8, 200);
            if (sent < 0)
                CLog::Log(LOGLEVEL_ERROR, errno, "CSendThread::execute: Sending failed");                
            else if (sent != 8)
                CLog::Log(LOGLEVEL_ERROR,"CSendThread::execute: Sending failed. Wrote only %u bytes", sent);

            usleep(100000);
        }
        delete [] buf;    
    }
}

void CButlerPhone::CSendThread::onWork(void *sender) {
    CLog::Log(LOGLEVEL_DEBUG,"CSendThread::onWork");
    ((CSendThread*)sender)->signalWork();
}

void CButlerPhone::CSendThread::signalWork(void) {
    pthread_mutex_lock(&condition_mutex);
    pthread_cond_signal(&condition_cond);
    pthread_mutex_unlock(&condition_mutex);    
}

void CButlerPhone::CSendThread::terminate(void) {
    CLog::Log(LOGLEVEL_DEBUG,"CSendThread::terminate");
    pthread_cancel(thread);
    CLog::Log(LOGLEVEL_DEBUG,"CSendThread::terminate canceled");
    pthread_cond_signal(&condition_cond);
    pthread_join(thread, NULL);
    CLog::Log(LOGLEVEL_DEBUG,"CSendThread::terminate cleaned up");
}
