/*
    $Id: CStatusThread.cc 60 2009-04-05 18:32:06Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "CLog.h"
#include "CButlerPhone.h"

CButlerPhone::CStatusThread::CStatusThread(CButlerPhone *owner, usb_dev_handle *h_USBDevice) {
    this->owner = owner;
    this->h_USBDevice = h_USBDevice;

    if (0 != pthread_create( &thread, NULL, threadfunc, this))
        CLog::Log(LOGLEVEL_ERROR, errno, "CStatusThread: Create StatusThread failed");
}

void CButlerPhone::CStatusThread::receive(void) {
    owner->receive(data,size);
}

void *CButlerPhone::CStatusThread::threadfunc(void *ptr) {
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::threadfunc");
    return ((CStatusThread*)ptr)->execute();
}

void CButlerPhone::CStatusThread::error(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::error calls set_Connected(false)");
    owner->set_Connected(false, NULL);
}

void *CButlerPhone::CStatusThread::execute(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::execute");
    pthread_mutex_t *synchro_mutex = owner->getMutex();

    int r, r2, i, l, messageok, goodchecksum, err;
    uchar buffer[0x80], b2[2];

    while(1) {
   //read the status message(s)
        r = usb_control_msg(this->h_USBDevice, 0xc0, 0x04, 0x5002, 0x0000, (char*)buffer, 0x80, 60000);
        if (r >=0)
            CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::execute: usb_control_msg read %d", r);
        else
            err = errno;

        // if terminated then break;
        pthread_testcancel();

        if (r >= 0) {
            //read the length of the previous message (braindead)
            r2 = usb_control_msg(this->h_USBDevice, 0xc0, 0x04, 0x500d, 0x0000, (char*)b2, 2, 500);

            // if terminated then break;
            pthread_testcancel();

            if (r2 == 2) {
                r2 = b2[0]*256 + b2[1];
        
                if (r2 == r) { 
                    messageok = true;
                 }
                else {
                    messageok = false;
                    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::execute: wrong length in second message, expected %d received %d",r,r2);
                }
            }
            else {
                messageok = false;
                CLog::Log(LOGLEVEL_DEBUG,"CStatusThread::execute: second message not received");
            }
      
            if (messageok) {
                    CLog::LogHex(LOGLEVEL_DEBUG, "CStatusThread::execute: Full msg", buffer,r);
                i = 0;
                //check all possible status messages
                while (i < (r-1)) {
                    if (buffer[i] != 0xf0) {
                        CLog::Log(LOGLEVEL_DEBUG,"CStatusThread::execute: status doesn't start with F0");
                        break;
                    }
                    l = buffer[i+1];

                    data = new uchar[l-1];
                    memcpy(data,&buffer[i+2],(l-1)*sizeof(uchar));

                    goodchecksum = (buffer[i+l+1] == checksum(data, l-1));
          
                    if (!goodchecksum) {
                        CLog::LogHex(LOGLEVEL_INFO,"CStatusThread::execute: wrong checksum",&buffer[i+l+1],1);
                        if (owner->isIgnoreChecksum)
                             CLog::Log(LOGLEVEL_INFO,"CStatusThread::execute: checksum ignored");
                    }

                    pthread_testcancel();
                    if (goodchecksum || owner->isIgnoreChecksum) {
                        this->size = l-1;
                        pthread_mutex_lock(synchro_mutex);
                        pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, synchro_mutex);
                        receive();
                        //pthread_mutex_unlock(synchro_mutex);
                        pthread_cleanup_pop(1);
                    }
                    i += (l + 1);
                    delete [] data;
                }                   
            }
            else {
                        CLog::Log(LOGLEVEL_INFO,"CStatusThread::execute: message ignored");
            }
        } //r >=0
        else {
            CLog::Log(LOGLEVEL_INFO, err, "CStatusThread::execute: Status read failed");
            CLog::Log(LOGLEVEL_INFO, "CStatusThread::execute: Trying to reconnect.");

            pthread_testcancel();
            pthread_mutex_lock(synchro_mutex);
            pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, synchro_mutex);
            error();
            //pthread_mutex_unlock(synchro_mutex);
            pthread_cleanup_pop(1);

            CLog::Log(LOGLEVEL_DEBUG, "CStatusThread EXITING normaly.");
            pthread_exit(NULL); 
        }
    } //while
}

void CButlerPhone::CStatusThread::terminate(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread::terminate");
    pthread_cancel(thread);
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread: canceled");
    pthread_join(thread, NULL);
    CLog::Log(LOGLEVEL_DEBUG, "CStatusThread: terminated");
}
