/*
    $Id: CSkypeAPI.cc 85 2009-08-31 01:38:17Z twimmer $; This file is part of:

    skiputler. Connects Topcom "Butler 4012 USB VoIP" with Skype on Linux.
    Copyright (C) 2009  Till Wimmer skiputler@tonarchiv.ch

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <errno.h>

#include <string>
#include <map>
#include <vector>


#include "CLog.h"
#include "CSkypeAPI.h"
#include "CSkypeAPIClient.h"


const char *CSkypeAPI::callStat_str[] = {
    "RINGING",
    "ROUTING",
    "FAILED",
    "INPROGRESS",
    "FINISHED",
    "MISSED",
    "REFUSED",
    "BUSY",
    "CANCELLED"
};

const char *CSkypeAPI::callType_str[] = {
    "INCOMING_PSTN",
    "INCOMING_P2P",
    "OUTGOING_PSTN",
    "OUTGOING_P2P"
};

CSkypeAPI::CSkypeAPI(CSkypeAPIClient *owner) {
    this->owner = owner;
    connection = NULL;
    proxy_service = NULL;
    proxy = NULL;

    threadConnect = NULL;
    threadRegister = NULL;
    threadNotify = NULL;
    threadPing = NULL;

    isConnected = false;
    isRegistered = false;
    ongoingCallId[0] = '\0';
    
    contactsIter = contacts.begin();

    pthread_mutex_init(&synchro_mutex, NULL);
    pthread_mutex_init(&invoke_mutex, NULL);

    pthread_mutex_init(&this->condition_mutex, NULL);
    pthread_cond_init(&this->condition_cond, NULL);

    g_type_init();
    g_thread_init(NULL);
}

CSkypeAPI::~CSkypeAPI() {
    if (threadPing) {
        threadPing->terminate();
        delete threadPing;
    }

    if (threadNotify) {
        threadNotify->terminate();
        delete threadNotify;
    }

    if (threadReceive) {
        threadReceive->terminate();
        delete threadReceive;
    }

    if (threadRegister) {
        threadRegister->terminate();
        delete threadRegister;
    }

    if (threadConnect) {
        threadConnect->terminate();
        delete threadConnect;
    }

    if (proxy) {
        g_signal_handler_disconnect(G_OBJECT(this->proxy), g_handler_id);
        g_object_unref(proxy);
    }

    if (proxy_service)
        g_object_unref(proxy_service);

    if (connection)
        dbus_g_connection_unref(connection);
}

// Callback function for handling DBusProxy destroy signal
void CSkypeAPI::destroy_signal(GObject *sender, gpointer data) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::DBusGProxy: 'destroy' signal triggered");

    ((CSkypeAPI*)data)->proxyDestroyed();
}

void CSkypeAPI::start() {
    GError *error;
    error = NULL;

    this->connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
    if (this->connection == NULL) {
        CLog::Log(LOGLEVEL_ERROR, "CSkypeAPI: Failed to open connection to bus: %s", error->message);
        g_error_free (error);
        exit(-1);
    }

    /* Create a proxy object for the "bus driver" (name "org.freedesktop.DBus") */
    this->proxy_service = dbus_g_proxy_new_for_name(this->connection,
                                            DBUS_SERVICE_DBUS,
                                            DBUS_PATH_DBUS,
                                            DBUS_INTERFACE_DBUS);

    threadReceive = new CSkypeReceiveThread(this);
    threadReceive->start();

    threadNotify = new CSkypeNotifyThread(this, this->connection);
    threadNotify->start();

    threadConnect = new CSkypeConnectThread(this, this->connection, this->proxy_service);
}

char *CSkypeAPI::invoke(const char *command) {
    char *reply = NULL;
    pthread_mutex_lock(&this->invoke_mutex);
    pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, &this->invoke_mutex);
    reply = this->_invoke(command);
    //pthread_mutex_unlock(&this->invoke_mutex);
    pthread_cleanup_pop(1);
    return reply;
}

char *CSkypeAPI::_invoke(const char *command) {
    GError *error = NULL;
    char *reply = NULL;

    if (this->proxy == NULL)
        return reply;

    proxycall_struct callstruct;

    callstruct.error = &error;
    callstruct.proxy = this->proxy;
    callstruct.command = command;
    callstruct.reply = &reply;
    callstruct.conditionMet = false;
    callstruct.condition_cond = &this->condition_cond;
    callstruct.condition_mutex = &this->condition_mutex;

    int ret = pthread_mutex_lock( &this->condition_mutex );
    if (ret) {
        CLog::Log(LOGLEVEL_ERROR, ret, "CSkypeAPI::invoke(): pthread_mutex_lock");
        return reply;
    }

    pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, &this->condition_mutex);

    // call Invoke, wait for reply
    g_idle_add(gloop_Invoke_call, &callstruct);

    while (!callstruct.conditionMet) {
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::invoke(): pthread_cond_wait...");
        pthread_cond_wait( &this->condition_cond, &this->condition_mutex );
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::invoke(): conditionMet = %d", callstruct.conditionMet);
    }

    //pthread_mutex_unlock( &this->condition_mutex );
    pthread_cleanup_pop(1);

    if (!callstruct.ret) {
        if (error) {
            CLog::Log(LOGLEVEL_ERROR, "CSkypeAPI::invoke(): dbus_g_proxy_call: %s", error->message);
            g_error_free (error);
        }
        return reply;
    }

    if (reply)
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::invoke(): -> \"%s\" , <- %s", command, reply);
    else
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::invoke(): -> \"%s\" , <- (null)", command);

    return reply;
}

void  CSkypeAPI::setConnected(bool status, DBusGProxy *proxy) {
    if (isConnected == status)
        return;

    isConnected = status;

    if(isConnected) {
        this->proxy = proxy;

        if (threadPing) {
            threadPing->terminate();
            delete threadPing;
            threadPing = NULL;
        }

        if (threadRegister) {
            threadRegister->terminate();
            delete threadRegister;
            threadRegister = NULL;
        }
        threadRegister = new  CSkypeRegisterThread(this);

        this->g_handler_id = g_signal_connect(
                    G_OBJECT(this->proxy),
                    "destroy",
                    G_CALLBACK(destroy_signal),
                    this);
    }
    else {
        owner->onSkypeOffline();

        isRegistered = false;
        
        g_signal_handler_disconnect(G_OBJECT(this->proxy), g_handler_id);

        objectunref_struct callstruct;

        callstruct.obj = this->proxy;
        callstruct.condition_cond = &this->condition_cond;
        callstruct.condition_mutex = &this->condition_mutex;

        //pthread_mutex_lock( &this->condition_mutex );
        //pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, &this->condition_mutex);

        //g_idle_add(gloop_objectUnref_call, &callstruct);
        g_object_unref(this->proxy);
        //pthread_cond_wait( &this->condition_cond, &this->condition_mutex );

        //pthread_mutex_unlock( &this->condition_mutex );
        //pthread_cleanup_pop(1);
        this->proxy = NULL;

        if (threadConnect) {
            threadConnect->terminate();
            delete threadConnect;
            threadConnect = NULL;
        }
        threadConnect = new CSkypeConnectThread(this, this->connection, this->proxy_service);
    }
}
void  CSkypeAPI::setRegistered(bool status) {
    if (isRegistered == status)
        return;

    isRegistered = status;

    if (isRegistered) {
        if (threadConnect) {
            threadConnect->terminate();
            delete threadConnect;
            threadConnect = NULL;
        }

        if (threadPing) {
            threadPing->terminate();
            delete threadPing;
            threadPing = NULL;
        }
        threadPing = new CSkypePingThread(this);
    }
    else {
        owner->onSkypeOffline();

        if (threadRegister) {
            threadRegister->terminate();
            delete threadRegister;
            threadRegister = NULL;
        }
        threadRegister = new  CSkypeRegisterThread(this);       
    }
}
void CSkypeAPI::proxyDestroyed(void) {

    if (threadPing) {
        threadPing->terminate();
        delete threadPing;
        threadPing = NULL;
    }

    if (threadRegister) {
        threadRegister->terminate();
        delete threadRegister;
        threadRegister = NULL;
    }

    if (threadConnect) {
        threadConnect->terminate();
        delete threadConnect;
        threadConnect = NULL;
    }

    pthread_mutex_lock(&this->synchro_mutex);
    pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, &this->synchro_mutex);
    setConnected(false, NULL);
    //pthread_mutex_unlock(&this->synchro_mutex);
    pthread_cleanup_pop(1);
}


pthread_mutex_t  *CSkypeAPI::getMutex() {
    return &this->synchro_mutex;
}

void CSkypeAPI::notify(const char *message) {
    if(message)
        this->threadReceive->receive(message);
}

void CSkypeAPI::_receive(const char *message) {
    //std::string strMessage(message);

    pthread_mutex_lock(&this->synchro_mutex);
    pthread_cleanup_push((void (*)(void*))pthread_mutex_unlock, &this->synchro_mutex);
    if (0==strncmp("USERSTATUS", message, 10)) {
        if (0==strncmp(" OFFLINE", &message[10], 7)) {
            owner->onSkypeOffline();
        }
        else {
            owner->onSkypeOnline();
        }
    }

    if (0==strncmp("CALL", message, 4))
        manageCall(message);

    //pthread_mutex_unlock(&this->synchro_mutex);
    pthread_cleanup_pop(1);
    CLog::Log(LOGLEVEL_DEBUG, "NOTIFY: \"%s\"", message);
}

void CSkypeAPI::manageCall(const char *message) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::manageCall(%s)", message);
    char *str, *ptr[5];
    int i=0;
    char msgbuf[128];

    strncpy(msgbuf, message, 128);
    
    ptr[i++] = strtok_r(msgbuf, " ", &str);

    while (i<5) {
        ptr[i] = strtok_r(NULL, " ", &str);
        if (!ptr[i])
            break;
        i++;
    }

    if (i < 3) {
        CLog::Log(LOGLEVEL_ERROR, "CSkypeAPI::manageCall(\"%s\") incomplete string", message);
        return;
    }

    char *callid = ptr[1];
    char *property = ptr[2];
    char *value = ptr[3];
    callStat_t callStatus = CSkypeAPI::callStat_UNDEF;

    CLog::Log(LOGLEVEL_INFO, "CALL Id=%s property=%s value=%s", callid, property, value);
    strncpy(this->ongoingCallId, callid, 16);

    if (0==strcmp("STATUS", property)) {
        for (unsigned int status=0; status < SIZE_ENUM_CALL_STAT; status++) {
            if (0==strcmp(CSkypeAPI::callStat_str[status], value)) {
                callStatus = (callStat_t)status;
                CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::manageCall: status = %s", callStat_str[callStatus]);
                break;
            }
        }
        
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::manageCall: call CSkypeAPIClient::onSkypeCall");
        owner->onSkypeCall(callStatus, getCallType(callid), getCallerName(callid));
    }

    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::manageCall end.");
    return;

}

char *CSkypeAPI::getCallerName(const char *callId) {
    std::string id(callId);
    std::string command;
    // TODO memory leaks
    command = "GET CALL " + id + " PARTNER_HANDLE";
    char *partner_handle = invoke(command.c_str());
    CLog::Log(LOGLEVEL_DEBUG, "partner_handle = \"%s\"", &partner_handle[command.size()-3]);

    command = "GET CALL " + id + " PARTNER_DISPNAME";
    char *partner_dispname = invoke(command.c_str());
    CLog::Log(LOGLEVEL_DEBUG, "partner_dispname = \"%s\"", &partner_dispname[command.size()-3]);

    return &partner_dispname[command.size()-3];
}

CSkypeAPI::callStat_t CSkypeAPI::getCallStat(const char *callId) {
    std::string id(callId);
    std::string command;
    callStat_t callStatus = CSkypeAPI::callStat_UNDEF;
    
    command = "GET CALL " + id + " STATUS";
    char *result = invoke(command.c_str());
    char *call_stat = &result[command.size()-3];
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::getCallStat(%s) = \"%s\"", callId, call_stat);

    for (unsigned int status=0; status < SIZE_ENUM_CALL_STAT; status++) {
        if (0==strcmp(CSkypeAPI::callStat_str[status], call_stat)) {
            callStatus = (callStat_t)status;
            break;
        }
    }

    free(result);
    return callStatus;
}

CSkypeAPI::callType_t CSkypeAPI::getCallType(const char *callId) {
    std::string id(callId);
    std::string command;
    callType_t callType = CSkypeAPI::callType_UNDEF;
    
    command = "GET CALL " + id + " TYPE";
    char *result = invoke(command.c_str());
    char *call_type = &result[command.size()-3];
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::getCallType(%s) = \"%s\"", callId, call_type);

    for (unsigned int type=0; type < SIZE_ENUM_CALL_TYPE; type++) {
        if (0==strcmp(CSkypeAPI::callType_str[type], call_type)) {
            callType = (callType_t)type;
            break;
        }
    }

    free(result);
    return callType;
}

time_t CSkypeAPI::getCallTimestamp(const char *callId) {
    std::string id(callId);
    std::string command;
    // TODO memory leaks
    command = "GET CALL " + id + " TIMESTAMP";
    char *timestamp = invoke(command.c_str());
    CLog::Log(LOGLEVEL_DEBUG, "timestamp = \"%s\"", &timestamp[command.size()-3]);

    return atoll(&timestamp[command.size()-3]);
}

void CSkypeAPI::initiateCall(const char *target) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::initiateCall(\"%s\")", target);
    std::string handle(target);
    std::string command("CALL ");

    command += target;

    char *result = invoke(command.c_str());
    // TODO
    free(result);    
}

void CSkypeAPI::initiateCallByName(const char *target) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::initiateCallByName(\"%s\")", target);
    std::map<std::string,std::string>::iterator iter;

    for( iter = this->contacts.begin(); iter != this->contacts.end(); ++iter ) {
        if (iter->second == std::string(target)) {
            initiateCall(iter->first.c_str());
            return;
        }
    }
    CLog::Log(LOGLEVEL_ERROR, "CSkypeAPI::initiateCallByName(\"%s\"): USER HANDLE not found.", target);        
}

void CSkypeAPI::acceptCall(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::acceptCall");
    std::string id(this->ongoingCallId);
    std::string command;

    command = "SET CALL " + id + " STATUS INPROGRESS";

    char *result = invoke(command.c_str());
    // TODO
    free(result);
}

void CSkypeAPI::hangupCall(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::hangupCall");
    std::string id(this->ongoingCallId);
    std::string command;

    command = "SET CALL " + id + " STATUS FINISHED";

    char *result = invoke(command.c_str());
    // TODO
    free(result);
}

void CSkypeAPI::updateContactList(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::updateContactList");
    std::string id(this->ongoingCallId);
    std::string command;
    std::vector<std::string> offline;
    std::map<std::string,std::string>::iterator iter;

    char *result, *str, *ptr;
    int count=0;
    
    this->contacts.clear();
    result = invoke("SEARCH FRIENDS");
    ptr = strtok_r(result, ", ", &str); // first token = "USER"

    while (count < SIZE_MAX_USERLIST) {
        ptr = strtok_r(NULL, ", ", &str);
        if (!ptr)
            break;
    
        this->contacts[std::string(ptr)] = "";
        count++;
    }

    free(result);

    //get offline contacts
    for( iter = this->contacts.begin(); iter != this->contacts.end(); ++iter ) {
        command = "GET USER " + iter->first + " ONLINESTATUS";
        result = invoke(command.c_str());
        if ( (0 == strcmp("OFFLINE", &result[command.size()-3]))
            || (0 == strcmp("UNKNOWN", &result[command.size()-3])) ) {
            offline.push_back(iter->first);
            CLog::Log(LOGLEVEL_DEBUG, "OFFLINE userhandle = %s", iter->first.c_str());            
        }
        free(result);
    }

    //delete offline contacts
    for (unsigned int i=0; i < offline.size(); i++)
        this->contacts.erase(offline[i]);

    // get FULLNAME
    for( iter = this->contacts.begin(); iter != this->contacts.end(); ++iter ) {
        command = "GET USER " + iter->first + " FULLNAME";
        result = invoke(command.c_str());
        if ( strlen(&result[command.size()-3]) )
            iter->second = std::string(&result[command.size()-3]);
        else
            iter->second = iter->first;
        free(result);
        CLog::Log(LOGLEVEL_DEBUG, "userhandle = %s, fullname = \"%s\"", iter->first.c_str(), iter->second.c_str());
    }

    // get DISPLAYNAME
    for( iter = this->contacts.begin(); iter != this->contacts.end(); ++iter ) {
        command = "GET USER " + iter->first + " DISPLAYNAME";
        result = invoke(command.c_str());
        if ( strlen(&result[command.size()-3]) )
            iter->second = std::string(&result[command.size()-3]);
        free(result);
        CLog::Log(LOGLEVEL_DEBUG, "userhandle = %s, displayname = \"%s\"", iter->first.c_str(), iter->second.c_str());
    }

}

const char *CSkypeAPI::getContactsFirst(void) {
    updateContactList();
    if(this->contacts.empty())
        return NULL;

    this->contactsIter = this->contacts.begin();
    
    return this->contactsIter->second.c_str();
}

const char *CSkypeAPI::getContactsNext(void) {
    if(this->contacts.empty())
        return NULL;

    this->contactsIter++;
    
    if(this->contactsIter == this->contacts.end())
        this->contactsIter = this->contacts.begin();

    return this->contactsIter->second.c_str();
}

const char *CSkypeAPI::getContactsPrevious(void) {
    if(this->contacts.empty())
        return NULL;

    if(this->contactsIter == this->contacts.begin())
        this->contactsIter = this->contacts.end();

    this->contactsIter--;
    
    return this->contactsIter->second.c_str();
}

void CSkypeAPI::updateCallsList(void) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::updateCallsList");

    std::string command;
    callsList_struct data;
    std::map<std::string,callsList_struct>::iterator iter;

    char *result, *handle, *ptr, *str;
    int count=0;
    
    this->calls.clear();
    result = invoke("SEARCH MISSEDCALLS");
    ptr = strtok_r(result, ", ", &str); // first token = "CALLS"

    while (count < SIZE_MAX_CALLSLIST) {
        ptr = strtok_r(NULL, ", ", &str);
        if (!ptr)
            break;

        handle = getCallerName(ptr);
        data.online = true;
        data.handle = std::string(handle);
        data.timestamp = getCallTimestamp(ptr);
        //free(handle);
        this->calls[std::string(ptr)] = data;
        count++;
    }

    free(result);

    //get offline contacts
    for( iter = this->calls.begin(); iter != this->calls.end(); ++iter ) {
        command = "GET USER " + iter->second.handle + " ONLINESTATUS";
        result = invoke(command.c_str());
        if ( (0 == strcmp("OFFLINE", &result[command.size()-3]))
            || (0 == strcmp("UNKNOWN", &result[command.size()-3])) ) {
            iter->second.online = false;
            CLog::Log(LOGLEVEL_DEBUG, "OFFLINE userhandle = %s", iter->second.handle.c_str());            
        }
        free(result);
    }
}

CSkypeAPI::callsList_struct *CSkypeAPI::getCallsFirst(void) {
    updateCallsList();
    if(this->calls.empty())
        return NULL;

    this->callsIter = this->calls.begin();
    
    return &this->callsIter->second;
}

CSkypeAPI::callsList_struct *CSkypeAPI::getCallsNext(void) {
    if(this->calls.empty())
        return NULL;

    this->callsIter++;
    
    if(this->callsIter == this->calls.end())
        this->callsIter = this->calls.begin();

    return &this->callsIter->second;
}

CSkypeAPI::callsList_struct *CSkypeAPI::getCallsPrevious(void) {
    if(this->calls.empty())
        return NULL;

    if(this->callsIter == this->calls.begin())
        this->callsIter = this->calls.end();

    this->callsIter--;
    
    return &this->callsIter->second;
}

gboolean CSkypeAPI::gloop_Invoke_call(gpointer data) {
    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::gloop_Invoke_call");
    proxycall_struct *callstruct =  (proxycall_struct *)data;
    char *reply, *buf, *outmarker;
    char *inmarker;
    
    size_t size, inlen, outlen;
    iconv_t cd;

    int ret = pthread_mutex_trylock(callstruct->condition_mutex);
    
    if (ret == EBUSY) {
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::gloop_Invoke_call condition_mutex BUSY");
        return TRUE; // do not unregester = retry later
    }

    if ( ret != 0) {
        CLog::Log(LOGLEVEL_ERROR, ret, "CSkypeAPI::gloop_Invoke_call");
        return FALSE;
    }
    
    callstruct->ret = dbus_g_proxy_call(callstruct->proxy, "Invoke", callstruct->error,
                            G_TYPE_STRING, callstruct->command, G_TYPE_INVALID,
                            G_TYPE_STRING, &reply, G_TYPE_INVALID);

    if (reply && (size = strlen(reply)) > 0) {
        inlen = outlen = size;

        inmarker = reply;
        if ((buf = (char *)malloc((size+1) * sizeof(char))) == NULL) {
            CLog::Log(LOGLEVEL_ERROR, errno, "CSkypeAPI::gloop_Invoke_call: malloc() error");
            goto cleanup;
        }
        outmarker = buf;
        
        if ((cd = iconv_open("ISO-8859-1", "UTF-8")) == (iconv_t)(-1)) {
            CLog::Log(LOGLEVEL_ERROR, errno, "CSkypeAPI::gloop_Invoke_call: iconv_open() error");
            goto cleanup;
        }

        // Convert every character within input string.
        while (inlen > 0) {
            if (iconv(cd, &inmarker, &inlen, &outmarker, &outlen) == (size_t)(-1)) {
                CLog::Log(LOGLEVEL_ERROR, errno, "CSkypeAPI::gloop_Invoke_call: iconv_string() conversion error");
                iconv_close(cd);
                goto cleanup;
            }
        }

        *outmarker = '\0';
        iconv_close(cd);
        *callstruct->reply = buf;
        free(reply);
    }
    
cleanup:
    callstruct->conditionMet = true;
    
    pthread_cond_signal(callstruct->condition_cond);
    pthread_mutex_unlock(callstruct->condition_mutex);

    CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::gloop_Invoke_call end.");
    return FALSE; // unregister from main loop
}

gboolean CSkypeAPI::gloop_objectUnref_call(gpointer data) {
    objectunref_struct *callstruct =  (objectunref_struct *)data;

    int ret = pthread_mutex_trylock(callstruct->condition_mutex);
    
    if (ret == EBUSY) {
        CLog::Log(LOGLEVEL_DEBUG, "CSkypeAPI::gloop_objectUnref_call condition_mutex BUSY");
        return TRUE; // do not unregester = retry later
    }

    if ( ret != 0) {
        CLog::Log(LOGLEVEL_ERROR, ret, "CSkypeAPI::gloop_objectUnref_call");
        return FALSE;
    }

    g_object_unref(callstruct->obj);

    pthread_cond_signal(callstruct->condition_cond);
    pthread_mutex_unlock(callstruct->condition_mutex);

    return FALSE; // unregister from main loop
}
